/* clwp 1.3: clwpasm.linux.s
   ASM code for the co-operative lwp package (Linux version)
   C.M. (Kees) Moerman, June 2002 - Januari 2003 (c)
   for documentation, see file clwp.txt
   Based on: lwp 2.0 of Josh Turpen

   Version 1.0: changed to co-operative
   Version 1.1: bug fix in _clwpDeadYield
   Version 1.3: stack overflow checking etc in clwpYield, clwpThrowAss added
*/

.globl clwpYield
.globl _clwpDeadYield
.globl _clwpThrowAss
.globl _clwpInitFpu
.globl _clwpInitFlags
.globl _clwpInitDs

/* offset in clwp struct definition for readability: */

/* typedef struct clwpStruct */
.set cScheck,      0               /* +0:  check value for validity */
.set cSpnext,      4               /* +4:  Pointer to next struct */
.set cSstklen,     8               /* +8:  Size of stack for this clwp */
.set cSpstackptr, 12               /* +12: Pointer to the stack */
.set cSpriority,  16               /* +16: Priority of clwp */
.set cSpcount,    20               /* +20: Used by scheduler */
.set cSactive,    24               /* +0x18: Thread suspend */
.set cSpstack,    28               /* +28: Start of stack (low mem) */
.set cSclwpid,    32               /* +32: Process ID */        


/* error message values used in assembly */
.set CLWPerrclwpnoinit,    31      /* clwp not initialised */
.set CLWPerrstackoverflow,  3      /* (clwp......) Stack damaged */
.set CLWPerrdamaged,        5      /* (clwp...) Internal data struct damaged */

/* special check values */

.set CLWPCHECK, 0x12345678         /* data structure marker */
.set CLWPSTACK, 0x5a5a5a5a         /* stack overflow marker */

.text

/* the main Yield routine: transfer control to next thread */

.align 4
.type  clwpYield,@function
clwpYield:                              /* yield control to next thread */
        pushal                          /* unless error detected */
        pushfl

        subl $108, %esp                 /* also save FPU state */ 
        fwait
        fnsave (%esp)                   /* FPU pushal */
        fwait

        movl _clwpCur, %esi             /* pointer to current task */
        cmpl $0, %esi                   /* may not be NULL !! (clwp not init'd) */
        je   yle
        
        cmpl $CLWPCHECK, cScheck(%esi)  /* should find special marker */
        jne  yld                        /* otherwise, data structure is damaged */
        
        movl %esp, cSpstackptr(%esi)    /* save stack pointer */
        
        movl cSpstack(%esi), %edi       /* need to get stack overflow marker */
        cmpl $CLWPSTACK,(%edi)          /* (tmpclwpPtr->stack[0] != 0x5A5A5A5A) */
        jne  yls                        /* stack overflow... */

yl1:    movl cSpnext(%esi), %esi        /* Find next ACTIVE thread */
        cmpl $0, cSactive(%esi)
        je   yl1

yl2:    decl cSpcount(%esi)             /* decr. priority counter */
        jnz  yl1                        /* not yet time for this task... */
        movl cSpriority(%esi),%eax      /* initialise counter with */
        movl %eax,cSpcount(%esi)        /* priority value */

yl3:    movl cSpstackptr(%esi), %esp    /* restore new thread's stack pointer */
        movl %esi, _clwpCur             /* save pointer for current thread */      
        cmpl $CLWPCHECK, (%esi)         /* ->check needs special marker */
        jne  yld                        /* otherwise, data structure is damaged */

ylr:    fwait                           /* and restore FPU state */
        frstor (%esp)                   /* FPU popal ;) */
        fwait
        addl $108, %esp

        popfl
        popal
        ret                             /* now we are in next thread */


/* major error situations: try to recover at least to main program */

yle:    movl $CLWPerrclwpnoinit, clwpLastError
        movl $-1, _clwpLastThread
        jmp  ylr                        /* return to main task */
        
yld:    movl $CLWPerrdamaged, clwpLastError
        movl cSclwpid(%esi), %eax       /* get thread id (pid) */
        movl %eax, _clwpLastThread
        jmp  yt1                        /* 'jump' to main task */
        
yls:    movl $CLWPerrstackoverflow, clwpLastError
        movl cSclwpid(%esi), %eax       /* get thread id (pid) */
        movl %eax, _clwpLastThread
        jmp  yt1                        /* 'jump' to main task */


/* Error handler kernel routine: go back to main thread */

.align 4
.type  _clwpThrowAss,@function
_clwpThrowAss:                          /* as yield, but to main thread */
        pushal
        pushfl

        subl $108, %esp                 /* also save FPU state */ 
        fwait
        fnsave (%esp)                   /* FPU pushal ;) */
        fwait

yt0:    movl _clwpCur, %esi             /* low level entries: */
        movl %esp, cSpstackptr(%esi)    /* save stack pointer */
yt1:    movl _clwpMain, %esi            /* and activate main thread */
        jmp  yl3                        /* continue there */        


/* Kill self */

.align 4
.type  _clwpDeadYield,@function
_clwpDeadYield:                         /* yield while killing self */
        decl _clwpCount
        movl _clwpCur, %esi
        movl %esi, %edi                 /* save for linked list compare */
        movl %esi, %ebx                 /* save for linked list walk */
        movl %esi, %edx                 /* clwp * to free */

ldy1:   movl %ebx, %ecx                 /* find previous thread */
        movl cSpnext(%ebx), %ebx
        cmpl %ebx, %edi
        jne  ldy1
        /* now %ecx = thread before clwpCur */

        movl cSpnext(%esi), %edi        /* remove clwpCur from list */
        movl %edi, cSpnext(%ecx)

ldy2:   movl cSpnext(%esi), %esi
        cmpl $0, cSactive(%esi)         /* find next ACTIVE thread */
        je   ldy2

        decl cSpcount(%esi)             /* decr. priority counter */
        jnz  ldy2                       /* not yet time for this task... */
        movl cSpriority(%esi),%eax      /* initialise counter with */
        movl %eax,cSpcount(%esi)        /* priority value */

        movl cSpstackptr(%esi), %esp
        movl %esi, _clwpCur 

        movl _clwpFreeList, %eax        /* get current free list ptr */
        movl %eax, cSpnext(%edx)        /* and store in 'next' field */
        movl %eax, (%edx)               /* clean check value */
        movl %edx, _clwpFreeList        /* this struct head of new list */

        fwait 
        frstor (%esp)                   /* FPU popal ;) */
        fwait
        addl $108, %esp

        popfl
        popal
        ret

.type  _clwpInitFpu,@function
_clwpInitFpu:                           /* retrieve default FPU state for init */
        pushl %eax
        movl $_clwpFpuState, %eax
        finit 
        fwait
        fnsave (%eax)
        fwait
        popl %eax
        ret

_clwpInitFlags:                         /* retrieve default flags for init */
.type  _clwpInitFlags,@function
        pushfl
        popl %eax
        ret

.type  _clwpInitDs,@function
_clwpInitDs:                            /* retrieve default segment for init */
        pushl %ds
        popl %eax
        ret

/* end of file */

