/* verification program for the clwp library version 1.3
   C.M. Moerman, Januari 2003 (c)
   
   This program runs several threads, verifying their execution sequence
   with a reference sequence. The sequence is stored as a string, each
   character indicating an executed task:
   0..9: main program cycle
   a: priority 1 task
   b: priority 2 task (run at half frequency)
   c: priority 4 task (quarter frequency)
   x: as 'a' but kills itself in cycle 5
   A: as 'a', however suspended several cycles (3-6)
   S: as 'a', however waits for semaphore (2-7)
   K: start kill sequence
   E: start error checking sequence
   O: stack overflow detection error (should not show up)
   ?: indicates unexpected status (should not show up)
   
   Note: this program also uses some symbols private to CLWP,
   esp. those using an underscore. Do not use these directly in your programs!
*/

#include "clwp.h"
#include <stdio.h>
#include <string.h>

#define CHECKLEN 250
char *expected = "0xSAa1xSAba2xAa3xcba4xa5ba6a7Acba8SAa9SAbaKSASAcbSASASASASSEb";
char result[CHECKLEN+4];
int iresult = 0;

static void add_char(char next)     /* add curent state to result */
{
    if(iresult != CHECKLEN)         /* check end of buffer space */
    {
/*      printf("%c", next);     */  /* useful when crashing before result output */
/*      fflush(stdout);         */
        result[iresult++] = next;   /* add character and trailing zero */
        result[iresult] = '\0';
    }
}

static int verify(void)             /* verify result with expectation */
{                                   /* returning error code */
    if(strcmp(expected, result))
    {
        printf("Mismatch,\nexpected: %s\n  result: %s\n", expected, result);
        return 1;
    }
    else
    {
        printf("Result verified OK:\n%s\n", result);
        return 0;
    }
}


/* the threads called for testing purposes: */

void proc1(void *param)      /* simple task, just enter id in result */
{   while(1)                        /* started multiple times in parallel */
    {   add_char( *((char *) param));
        clwpYield();                /* switch to next thread */
    }
}

int mutex;                          /* semaphore test mutex */

void proc_sema(void *param)         /* this proc checks simple semaphores */
{   while(1)
    {   clwpLockMutex(&mutex);      /* lock and release semaphore */
        clwpReleaseMutex(&mutex);   /* maight also be locked by main */
        add_char( *((char *) param));
        clwpYield();                /* switch to next thread */
    }
}

void proc_kill(void *param)         /* this proc checks suiside */
{   while(1)
    {   
        if( *((int *) param) == 5) clwpKill(clwpGetpid());
        add_char('x');
        clwpYield();                /* switch to next thread */
    }
}

/* the main program running the tests */

int main()                          /* run threads, verify sequence */
{
    int i;                          /* loop counter, also arg for proc_kill */
    int ids[6];                     /* to store thread IDs */

    char a='a',b='b',c='c',d='A',e='S';    /* thread arguments */

    printf("\nCLWP verification run. Ignore error messages,\n"
           "just look for 'Result verified OK' or 'Mismatch' message at end\n");
    fflush(stdout);

    if(clwpInit(1))                 /* initialise the threading system */
    {                               /* and create 3 simple threads */
        if(!clwpCreateMutex(&mutex))
            add_char('?');          /* Could not allocate mutex... */

        ids[0] = clwpSpawn(proc1, &a, 4096, 1, TRUE);
        ids[1] = clwpSpawn(proc1, &b, 4096, 2, TRUE);
        ids[2] = clwpSpawn(proc1, &c, 4096, 4, TRUE);
        ids[3] = clwpSpawn(proc1, &d, 4096, 1, TRUE);  /* will be suspended */
        ids[4] = clwpSpawn(proc_sema, &e, 4096, 1, TRUE);  /* sema wait */
        ids[5] = clwpSpawn(proc_kill, &i, 4096, 1, TRUE);  /* suicide */
                                    /* should all be positive */
        if(ids[0]*ids[1]*ids[2]*ids[3]*ids[4]*ids[5]<=0) add_char('?');
        if(clwpCheck()) add_char('?');

        for(i=0;i<10;i++)           /* go looping, activating all threads */
        {
            add_char('0'+i);        /* 0..9 index in result */

            if(i==3) clwpThreadSuspend(ids[3]);
            if(i==7) clwpThreadResume(ids[3]);

            if(i==2) clwpLockMutex(&mutex);
            if(i==8) clwpReleaseMutex(&mutex);
            
            clwpYield();
        }
        
        if(clwpCheck()) add_char('?');
        
        add_char('K');              /* start killing */
        clwpKill(ids[0]); clwpYield(); clwpYield();
        clwpKill(ids[1]); clwpYield(); clwpYield();
        clwpKill(ids[2]); clwpYield(); clwpYield();
        clwpKill(ids[3]); clwpYield(); clwpYield();
        clwpKill(ids[4]); clwpYield(); clwpYield();
        
/* also check error handling/trapping: */
        add_char('E');              /* marker of start error checking */
        
        clwpDeleteMutex(&i);        /* this is not a mutex! */
        if(clwpLastError != CLWPerr_semanotexist) add_char('!');
        clwpThrow(0);

        if(!clwpDeleteMutex(&mutex)) 
            add_char('?');          /* Could not delete mutex...? */
        if(clwpLastError)           /* should still be clean */
        { add_char('?'); add_char('0'+clwpLastError%10); }
    
        clwpDeleteMutex(&i);        /* this is not a mutex! */
        if(clwpLastError != CLWPerr_semalistzero) add_char('!');

        clwpKill(clwpGetpid());     /* may not delete main! */
        if(clwpLastError != CLWPerr_cantkillmain) add_char('!');
        clwpKill(ids[0]);           /* is already dead! */
        if(clwpLastError != CLWPerr_pidnotfound) add_char('!');

/* Official error number retrieval: use clwpCatch() */
        if(clwpCatch() != (CLWPerr_pidnotfound | (-1 << 16))) add_char('C'); 

        clwpThrow(1234);            /* this is already main thread, should be OK */
        if(clwpLastError != 1234) add_char('T');
        if(clwpCatch() != (1234 | (-1 << 16))) add_char('T');
        
/* still to add: more destructive tests like stack overflow */ 
/* needs access to internal variables like handle to thread structure */
/* for that purpose: _clwp_gen_overflow in clwp.c */

        ids[0] = clwpSpawn(proc1, &a, 4096, 1, TRUE);
        ids[1] = clwpSpawn(_clwp_gen_overflow, &c, 4096, 1, TRUE);
        ids[2] = clwpSpawn(proc1, &b, 4096, 1, TRUE);
        clwpYield();                /* should generate overflow in ids[1] */
        if(clwpCatch() != (CLWPerr_stackoverflow | (ids[1] << 16)))
            add_char('O');          /* checks task number and error code */

    }
    
    clwpThrow(0);                   /* don't do this at home! */
    clwpExit();                     /* should also run without errors */
    if(clwpCatch() & 0xFFFF) add_char('?');
   
    return(verify());               /* verify and return error code */
}

/* end of file */
