/*
 * Pinguin game helper functions: helpers.c
 * C.M. Moerman, Februari 2002 - Februari 2003
 */

#define ALLEGRO_NO_COMPATIBILITY 1
#include "main.h"

/* some local 'forward' references */
int debug_on = 0;                       /* F11/F12 disable/enable debug */

int keyhandler_id = 0;                  /* CLWP task id for keyboard handler */
int watchdog = 0;                       /* watchdog on threading */

typedef struct FM_SPRITE {              /* image in display list */
    int x, y;
    int rotate, mirror;
    BITMAP * image;                     /* sprite bitmap to display */
    struct FM_SPRITE *next;             /* for linked lists */
    int check;                          /* just for security */
} FM_SPRITE;
#define FM_CHECK 4711                   /* used to verify valid node ID */
#define FM_FREE 4712                    /* released node ID */

/* ---------------------------------------------------------------------
 * Routines for bitmap allocation for score, help text etc.
 * --------------------------------------------------------------------- */

BITMAP * dbgbmpt = NULL;                /* For debug text message */
BITMAP * scorebmpt = NULL;              /* For points scored */
BITMAP * helpbmpt = NULL;               /* For help messages */
BITMAP * overlaybmpt = NULL;            /* various overlays like volume bar */
#define OVERLAY_W 256                   /* overlay for volume controls etc sizes */
#define OVERLAY_H 32

void init_buffers(void)
{
/* Allocate box for points scored text */
    scorebmpt = create_system_bitmap (100, 30);
    if (scorebmpt==NULL) { exit_buffers(); fm_assert(scorebmpt!=NULL); fm_exit(); exit (2); }
    clear_to_color (scorebmpt, makecol(255, 0, 255));

/* Allocate box for help message output */
    helpbmpt = create_system_bitmap (SWIDTH-100, SHEIGHT-200);
    if (helpbmpt==NULL) { exit_buffers(); fm_assert(helpbmpt!=NULL); fm_exit(); exit (2); }
    clear_to_color( helpbmpt, makecol(255, 0, 255));

/* Allocate box for debug message output */
    dbgbmpt = create_system_bitmap (SWIDTH-(LMARGIN+RMARGIN), 8);
    if (dbgbmpt==NULL) { exit_buffers(); fm_assert(dbgbmpt!=NULL); fm_exit(); exit (2); }
    clear_bitmap( dbgbmpt);

/* Allocate box for overlay output */
    overlaybmpt = create_system_bitmap (OVERLAY_W, OVERLAY_H);
    if (overlaybmpt==NULL) { exit_buffers(); fm_assert(overlaybmpt!=NULL); fm_exit(); exit (2); }
    clear_bitmap( overlaybmpt);
}

FM_SPRITE *dl_freelist = NULL;

void exit_buffers(void)
{   FM_SPRITE *temp;
    if(dbgbmpt != NULL) destroy_bitmap (dbgbmpt);
    if(helpbmpt != NULL) destroy_bitmap (helpbmpt);
    if(scorebmpt != NULL) destroy_bitmap (scorebmpt);
    if(overlaybmpt != NULL) destroy_bitmap (overlaybmpt);
    while (dl_freelist != NULL)           /* release empty block list */
    {   fm_assert(dl_freelist->check == FM_FREE);
        temp = dl_freelist->next; free(dl_freelist); dl_freelist = temp;
    }
    MapFreeMem ();
}


/* ---------------------------------------------------------------------
 * Routines for display list manipulation
 * --------------------------------------------------------------------- */

void displaylist_new(void **list, int total_items) /* get new display list */
{   int i;
    FM_SPRITE *new_dl;
    for(i=0; i<total_items; i++)
    {   new_dl = (FM_SPRITE *)malloc(sizeof(FM_SPRITE));
        fm_assert(new_dl != NULL);
        new_dl->next = dl_freelist;
        new_dl->check = FM_FREE;
        dl_freelist = new_dl;
    } 
    *list = NULL;
}

void displaylist_draw(void *list)      /* display list */
{   POINT scr;
    FM_SPRITE *temp = (FM_SPRITE *)list;
    int margin;
    
    while(temp != NULL)
    {
        fm_assert(temp->check == FM_CHECK);
        scr.x = (temp->x-mapoff.x)+LMARGIN;    /* translate to screen coordinates */
        scr.y = (temp->y-mapoff.y)+TMARGIN;    /* within displayable area? */
        margin = temp->rotate?8:0;             /* rotated sprite may extend */
        if((scr.x>margin) && (scr.x < SWIDTH - temp->image->w - margin)
        && (scr.y>margin) && (scr.y < SHEIGHT - temp->image->h - margin))
        {
            if(abs(temp->rotate) <= itofix(1)) /* if almost not rotated: take fast route */
            {                                  /* shouldn't we mirror the image? */
                if(!temp->mirror) draw_sprite(backbuffer, temp->image, scr.x, scr.y);
                else draw_sprite_h_flip(backbuffer, temp->image, scr.x, scr.y);
            }
            else
            { 
                if(!temp->mirror) rotate_sprite(backbuffer, temp->image, scr.x, scr.y, temp->rotate-itofix(32));
                else rotate_sprite_v_flip(backbuffer, temp->image, scr.x, scr.y, temp->rotate+itofix(128)+itofix(32) );
            }
        }
        temp = temp->next;
    }
}

void displaylist_clear(void **list)     /* clean list: add to empty list */
{
    FM_SPRITE *temp  = (FM_SPRITE *)*list;    /* casted list */
    FM_SPRITE *clist = (FM_SPRITE *)*list;  
    while( temp != NULL)
    {   fm_assert(temp->check == FM_CHECK);   /* data struct not damaged/overwritten? */
        temp->check = FM_FREE;
        temp = clist->next;
        clist->next = dl_freelist;
        dl_freelist = clist;
        clist = temp;
    }
    *list = NULL;
}

void displaylist_add(void **list, BITMAP *image, int x, int y, int mirror, int rotate)
{
    FM_SPRITE *temp = (FM_SPRITE *)*list;
    if(dl_freelist == NULL)                 /* no empty nodes left? */
    {   dl_freelist = malloc(sizeof(FM_SPRITE));
        fm_assert(dl_freelist != NULL);     /* request new node */
        dl_freelist->next = NULL;
        dl_freelist->check = FM_FREE;
    }    
    temp = dl_freelist;
    fm_assert(temp->check == FM_FREE);      /* data struct not damaged/overwritten? */
    temp->check = FM_CHECK;
    dl_freelist = dl_freelist->next;
    temp->next = *list;
    temp->x = x; temp->y = y; temp->image = image;
    temp->rotate = rotate; temp->mirror = mirror;
    *list = temp;                           /* update list ptr in calling program */
}

/* ---------------------------------------------------------------------
 * Routines for basic collision detection etc
 * --------------------------------------------------------------------- */

/* This is an extremely basic collision detection routine, a proper game would
 * need a better one
 */
BLKSTR * Collision (int x, int y)
{
BLKSTR * blkdatapt;

    blkdatapt = MapGetBlock (x/mapblockwidth, y/mapblockheight);

    if (blkdatapt->tl) return blkdatapt; else return NULL;
    // user5 = 1: eat/score points
    // user5 = 2: help message
    // user5 = 3: baby pinguin --> see main.h for more values
}

/* This simply removes an object once it's collected */
void ClearCell (int x, int y)
{
    MapSetBlock (x/mapblockwidth, y/mapblockheight, 0);
}

void SetCell (int x, int y, int block_id)
{
    MapSetBlock (x/mapblockwidth, y/mapblockheight, block_id);
}

BLKSTR * ConfigBlock (void)             /* get level configuration block */
{
    return MapGetBlock (0, 0);          /* see map_doc.txt for fields */
}


void log_state(void)                    /* for debug purposes: logging */
{    /* framer should be 0 or 1 except for initial image */
     char logline[256];
     sprintf(logline,
         "State info: framer = %d; fps = %d; debug = %d, %d, %d; level = %d; score = %d",
         fm_framer, fm_fps, debug1, debug2, debug3, level+1, score);
     log_text(2, logline);
}

/* ---------------------------------------------------------------------
 * Routines to update and display the help text and overlay frame
 * --------------------------------------------------------------------- */

int overlay_x = 0;                      /* overlay e.g. for volume control */
int overlay_y = 0;                      /* overlay_x also used as flag */

int help_active = 0;
char *phelptext = NULL;                 /* pointer to help text */
int oldhelp = -1;                       /* index to old help (refresh needed?) */

char *find_in_dat(int level, int index) /* find (N,M)'th block of help */
{                                       /* Note: level = -1: general help */
    char *helpptr;
    helpptr = phelptext;                /* pointer to help text from init_help */
    fm_assert( *helpptr == '#' );
    while (!(    ((*helpptr=='#') && (*(helpptr+1)=='!'))
              || ((*helpptr=='#') && (*(helpptr+1)==('A'+level)) && (*(helpptr+2)==('0'+index)))
          ))
        helpptr++;                      /* scan for '#<index>' */
    fm_assert( *helpptr == '#' );
    helpptr += 4;                       /* skip '#?? ' */
    return helpptr;
}

char *copy_line(char *dest, char *source)
{
    while (*source != 0x0D) *dest++ = *source++;
    *dest = 0;
    source++; while((*source == 0x0A) || (*source == 0x0D)) source++;
    return source;
}

void update_help(int level, int index)
{
    char *helpptr;
    int linecntr = 0;
    char linebuf[100];

    if( (index+100*level) != oldhelp)
    {
        log_num(3, "Help activated ", 1000+index);

        clear_to_color (helpbmpt, makecol(255, 0, 255));
        helpptr = find_in_dat(level, index);
        while(*helpptr != '#')
        {
            helpptr = copy_line(linebuf, helpptr);
            textout_centre(helpbmpt, data[FM_HFONT].dat, linebuf, (SWIDTH-100)/2, linecntr++*20, -1);
        }
        helpptr = find_in_dat(-1, 3);   /* #@3: press space to continue */
        helpptr = copy_line(linebuf, helpptr);
        textout_centre(helpbmpt, data[FM_HFONT].dat, linebuf, (SWIDTH-100)/2, linecntr*20+10, -1);

        oldhelp = index+100*level;
    }
    help_active = 1000;                 /* count down at player update rate */
}

void display_helptext(int xoffset, int yoffset)
{
    if(help_active > 0) draw_sprite(backbuffer, helpbmpt, xoffset, yoffset);
    if(overlay_x) draw_sprite(backbuffer, overlaybmpt, overlay_x, overlay_y);
}

void init_help(void)                    /* try to find external help file, */
{    int bytes_read;                    /* if not found use internal (Dutch) */
     FILE * handle;
     handle = fopen ("help.txt", "rb"); /* the help text file */

     if(handle == NULL)
         phelptext = (char *)data[FM_HELP1].dat; /* pointer from database */
     else
     {   phelptext = malloc(17000);
         fm_assert(phelptext != NULL);
         bytes_read = fread ( (void *)phelptext, 1, 16500, handle);
         fclose(handle);               /* file incorrect length? */
         if((bytes_read > 16384) || (bytes_read < 100))
         {   log_num( 1, "Invalid help length ", bytes_read);
             fm_assert(bytes_read <= 16384);
         }
         else                          /* just to be sure.. */
             strcpy(phelptext+bytes_read, "#!! End of help text\n\r\n\r");
         log_num( 2, "External help, length ", bytes_read);
     }    
}

void exit_help(void)                   /* release help text buffer */
{
     if((phelptext != NULL) && (phelptext != (char *)data[FM_HELP1].dat))
         free(phelptext);
     log_num(2, "Key thread stack used: ", clwpStackUsed(keyhandler_id));

}



/* ---------------------------------------------------------------------
 * Routines to handle menu keys etc
 * --------------------------------------------------------------------- */
 
int read_name_flag = 0;                 /* used in keyboard routine, */
                                        /* set in high score handling */
char new_name[128];

int midi_volume = 128;                  /* volume control */
int wave_volume = 160;

void create_rectangle(int value, char *text)
{                                       /* bar graph indicating volume settings etc */
    overlay_y = SHEIGHT/2; overlay_x = (SWIDTH-OVERLAY_W)/2;
    rectfill(overlaybmpt, 0, 0, OVERLAY_W-1, OVERLAY_H-1, makecol(128, 128, 128));
    rectfill(overlaybmpt, 0, 0, value, OVERLAY_H-1, makecol(255, 0, 0));
        rect(overlaybmpt, 0, 0, OVERLAY_W-1, OVERLAY_H-1, makecol(255, 255, 255));
    textout_centre(overlaybmpt, data[FM_HFONT].dat, text, OVERLAY_W/2, 8, -1);
}

void set_midi_volume(int flag)
{                                       /* adjust midi volume, display bar */
    midi_volume = min(max(midi_volume+flag, 0), 255);
    set_volume(-1, midi_volume);
    create_rectangle(midi_volume, "Midi Music");
}

void set_wave_volume(int flag)
{                                       /* adjust wave volume, display bar */
    wave_volume = min(max(wave_volume+flag, 0), 255);
    set_volume(wave_volume, -1);
    create_rectangle(wave_volume, "Wave effects");
}

void read_name(void)                   /* read name for high-score list */
{
    int i = 0;
    int answer;                        /* one-character input */
    char prompt[32];                   /* prompt string for text box */
    
    while( keypressed()) { readkey(); } /* empty keyboard input buffer */
    
    overlay_x = (SWIDTH-OVERLAY_W)/2; overlay_y = SHEIGHT/3;
    sprintf(prompt, "%5d: _______ !!", score);     /* 7 underlines */
    help_active = 10000;
    new_name[0] = '\0';
    while( i < 7 )
    {
         rectfill(overlaybmpt, 0, 0, OVERLAY_W-1, OVERLAY_H-1, makecol(128, help_active & 0xFF, 64));
         rect(overlaybmpt, 0, 0, OVERLAY_W-1, OVERLAY_H-1, makecol(255, 255, 255));
         rect(overlaybmpt, 1, 1, OVERLAY_W-2, OVERLAY_H-2, makecol(0, 0, 0));
         textout_centre(overlaybmpt, data[FM_HFONT].dat, prompt, OVERLAY_W/2, 8, -1);
         
         while( !keypressed())
         {
              clwpYield();              /* don't block other processes! */
              if(help_active > 0) help_active--;
              if(help_active < 0) help_active++;
              watchdog = 0;
         }
         answer = readkey();
         switch(answer >> 8)
         {
             case KEY_BACKSPACE: { if (i>0) { i--; prompt[7+i] = '_'; } break; }
             case KEY_ENTER:     { i = 100; break; }
             case KEY_ESC:       { i = 200; break; } 
             case KEY_LEFT:      { if (i>0) i--; break; }
             case KEY_RIGHT:     { if (i<7) i++; break; }
             default:
                 { if( ((answer&0xFF) > 0x1F) && ((answer&0xFF) < 0x7F))
                      { prompt[7 + i++] = answer & 0xFF; }
                 }
         }
         clwpYield();                   /* don't block other processes! */           
    }
    while( keypressed()) { readkey(); } /* empty keyboard input buffer */

    overlay_x = 0;
    help_active = -50;                  /* remove text from screen */
    read_name_flag = 0;                 /* ready reading name */
    if (i == 200) return;               /* escape typed -> don't update */
    
    memcpy(new_name, prompt+7, 7);      /* copy name */
    new_name[7] = '\0';
    for(i=6; i>=0; i--)                 /* find end of name */
        if(new_name[i] == '_') new_name[i] = '\0'; else i = -1;
    log_num(2, "Name read:", strlen(new_name));
    log_text(2, new_name);
    return;
}

int ynflag = 0;

void key_esc_yesno(void)                /* check yse/no, set quit accordingly */
{   char answer;

    if (ynflag == 0 && key[KEY_ESC])
    {   while( key[KEY_ESC] ) {};       /* wait for key release */
        update_help(-1, 2);             /* #@2: Quit: Yes or No */
        ynflag = 10;                    /* give a few frames to update display */
    }
    if(ynflag > 0) ynflag--;            /* assume text on display? */
    if(ynflag == 1)                     /* waiting for answer.... */
    {
        while( keypressed()) { readkey(); } /* empty keyboard input buffer */
        while( !keypressed()) { watchdog = 0; clwpYield(); }
        answer = readkey();
        if( (answer >> 8) == KEY_ESC) quit = 1;
        answer &= 0xDF;                 /* map to upper case */
        if((answer == 'Y') || (answer == 'J') || (answer == 'O')) quit = 1;
        ynflag = 0;
        help_active = -50;              /* remove text */
    }
}

void keyhandler_thread(void *param)     /* handle keys in separate thread */
{                                       /* except for carrow keys */
    while(1)
    {    
        if(help_active > 0) help_active--;
        if(help_active < 0) help_active++;

        if(read_name_flag) read_name(); /* don't return; will kill thread */

        key_esc_yesno();                /* check ESC key, check Yes/No, set quit */

        if (key[KEY_SPACE]) help_active = -50;     /* delay till next text */

        if (key[KEY_F1]) update_help(-1, 0); /* this is help #@0 for F1 */
        if (key[KEY_F4]) set_midi_volume(-1);
        if (key[KEY_F5]) set_midi_volume(+1);
        if (key[KEY_F6]) set_wave_volume(-1);
        if (key[KEY_F7]) set_wave_volume(+1);
        if ( !(key[KEY_F4] || key[KEY_F5] || key[KEY_F6] || key[KEY_F7])) overlay_x = 0;
    
        if (key[KEY_F8]) display_hiscore();
        
/*      if (key[KEY_F9]) cheat by not checking horizontal collisions: in move routines */
/*      if (key[KEY_F10]) score = 10000; */

        if (key[KEY_F11]) debug_on = 0;
        if (key[KEY_F12]) debug_on = 1;
    
        if (key[KEY_S])   snow_on = 0;  /* for speed increase on slow systems */
        if (key[KEY_V])   fish_on = 0;
    
        clwpYield();
        watchdog = 0;
    }
}

void init_keyhandler()                  /* set up thread for key handling */
{   
    if(keyhandler_id != 0) return;
    
    ynflag = 0;
    set_volume(wave_volume, midi_volume);
    keyhandler_id = clwpSpawn(keyhandler_thread, &keyhandler_id, 16384, 1, TRUE);
    fm_assert( keyhandler_id != 0);
}



/* ---------------------------------------------------------------------
 * Routines to update and display the score
 * --------------------------------------------------------------------- */

int score = 0;
int flash = 0;

void update_score(void)                /* everything related to scoring etc */
{
    static int oldscore = -1;

    if(score<0) score = 0;
    if(score != oldscore)
    {
        clear_to_color (scorebmpt, makecol(255, 0, 255));
        textprintf_centre(scorebmpt, data[FM_HFONT].dat, 100/2, 2, -1, "%d", score);
        oldscore = score;
    }

    if((airsupply<=0) && !(flash & 0x0F)) score--; /* check air supply */
    if((airsupply<=0) && !(flash & 0x1F)) play_sample(data[FM_AIRBEEP].dat, 128, 0, 1000, 0);
    flash++;
}

void display_score(void)                /* everything related to display score etc */
{
    int i;

    draw_sprite(backbuffer, data[FM_KADERL].dat, 0, 0);
    draw_sprite(backbuffer, data[FM_KADERR].dat, SWIDTH-32, 0);
    draw_sprite(backbuffer, data[FM_KADERT].dat, 32, 0);
    draw_sprite(backbuffer, data[FM_KADERB].dat, 32, SHEIGHT-32);

    for (i=0; i<4; i++)                 /* air bar: (185,14) .. (188,65) -> (4,51) */
    {                                   /* warning: low air at 20%*/
        vline(backbuffer, 185+i, 14, 65, makecol(64,0,0));
        vline(backbuffer, 185+i, 65-(airsupply*50/maxair), 65,
            makecol((airsupply < maxair/5)?255:0,  /* red or cyan? */
               (airsupply < maxair/5)?0:255, (airsupply < maxair/5)?0:255));
    }
    if(airsupply < maxair/5)
        draw_sprite(backbuffer,         /* green (no draw), red or blinking? */
            data[((flash & 0x08)||(airsupply>0))?FM_LRED:FM_LBLACK].dat, 201, 3);

    for(i=0; i<total_babies; i++)       /* display area: (500,35) .. (600, 66) */
        draw_sprite(backbuffer, data[FM_PDISP1+((saved_babies>i)?1:0)].dat,
            490+(120-32)*i/(level+1), 35);

    draw_sprite(backbuffer, scorebmpt, 40, 25);
}

#define CONFIG_LIST 5                   /* number of hi scores in list */
#define CONFIG_LEN (10+CONFIG_LIST*10+2) /* globals, 5 * hiscore (2+8), checksum */
#define CONFIG_NAME "pinpin.cfg"

char config[CONFIG_LEN];

int getword(char *wordptr)              /* read *(msb,lsb) into int */
{
    return ((*wordptr<<8) & 0xFF00) + (*(wordptr+1) & 0xFF);
}

void putword (char *wordptr, int word)  /* write int into *(msb,lsb) */
{
     *wordptr = (word>>8) & 0xFF; *(wordptr+1) = word & 0xFF;
}

void init_hiscore(void)                 /* in case of failure: re-init structure */
{    int i;
     for(i=0; i<CONFIG_LEN; i++) config[i] = 0;
     for(i=0; i<CONFIG_LIST; i++)
     { putword(&config[10+i*10], 1 <<(10-i)); strcpy(&config[10+i*10+2], "Kees"); }
}

int test_hiscore(void)                  /* new hi-score? */
{    /* compare against lowest hi score, true if higher */
     return (score > getword( &config[10 + (CONFIG_LIST-1)*10] ));
}

void save_hiscore(void)                 /* hiscore and some global settings */
{    int i, check;
     FILE * handle;
     
     if(strlen(new_name) == 0) return;  /* not needed: no new name */

     for (i=0; i<CONFIG_LIST; i++)
     {
         if(score > getword(&config[10+i*10]))
         {
             memmove(&config[10+(i+1)*10], &config[10+i*10], max(0, 10*(4-i)));
             putword(&config[10+i*10], score);
             strcpy (&config[10+i*10+2], new_name);
             i = CONFIG_LIST;           /* done, leave loop */
         }
     }

     check = 0; for(i=0; i<CONFIG_LEN-2; i++) check += config[i]; /* checksum */
     putword(&config[CONFIG_LEN-2], check);

     handle = fopen (CONFIG_NAME, "wb");
     if(handle == NULL)
     { log_text( 1, "Can not save hiscore"); return; }
     fwrite ( (void *)config, 1, CONFIG_LEN, handle);
     fclose(handle);
}

void load_hiscore(void)                 /* read or remake hiscore */
{    int bytes_read;
     int i, check;
     FILE * handle;
     handle = fopen (CONFIG_NAME, "rb"); /* the file */
     if(handle == NULL)
     {   log_text( 1, "Can not read hiscore");
         init_hiscore();                /* file does not exist.... */
     }
     else
     {
         bytes_read = fread ( (void *)config, 1, CONFIG_LEN, handle);
         fclose(handle);               /* file incorrect length? */
         if(bytes_read != CONFIG_LEN)
         {   log_text( 1, "Invalid hiscore length");
             init_hiscore();
         }
         else
         {
             check = 0; for(i=0; i<CONFIG_LEN-2; i++) check += config[i];
             check -= getword(&config[CONFIG_LEN-2]);
             if(check)
             {   log_text( 1, "Invalid hiscore");
                 init_hiscore();
             }
         }
     }
}

void display_hiscore(void)              /* hiscore list */
{
    char *helpptr;
    int linecntr = 0;
    char linebuf[100];
    int i;

    clear_to_color (helpbmpt, makecol(255, 0, 255));
    textout_centre(helpbmpt, data[FM_HFONT].dat, "  Hi Score", (SWIDTH-100)/2, linecntr++*20, -1);
    for(i=0; i<CONFIG_LIST; i++)
    {    sprintf(linebuf, "%s %5d", &config[10+i*10+2], getword(&config[10+i*10]));
         textout_centre(helpbmpt, data[FM_HFONT].dat, linebuf, (SWIDTH-100)/2, linecntr++*20, -1);
    }
    linecntr++;
    helpptr = find_in_dat(-1, 3);       /* #@3 Space to continue */
    helpptr = copy_line(linebuf, helpptr);
    textout_centre(helpbmpt, data[FM_HFONT].dat, linebuf, (SWIDTH-100)/2, linecntr*20+10, -1);

    help_active = 1000;                 /* count down at player update rate */
    oldhelp = -1;                       /* reset old bitmap */
}

/* ---------------------------------------------------------------------
 * Routines to update and display the debug information (frame rate etc)
 * --------------------------------------------------------------------- */

char debug_textbuf[128] = "Debug text";
int debug1 = 0;                         /* these fields appl.dependent */
int debug2 = 0;
int debug3 = 0;
int debug_timer = 0;

void display_debug_text(void)           /* some debug info */
{
    if(debug_on && !(debug_timer++ & 0x0F))  /* not too often (save time) */
    {   clear_bitmap( dbgbmpt);
        sprintf(debug_textbuf, "pos:%3d,%3d speed:%3d,%3d rotate:%3d fps:%3d hw=%d   d1=%d d2=%d d3=%d",
            (int)player.x, (int)player.y, (int)player.vx, (int)player.vy, fixtoi(player.rotate),
            fm_fps, fm_hwaccel, debug1, debug2, debug3);
        textout( dbgbmpt, font, debug_textbuf, 0, 0, makecol (0, 255, 255));
    }
    if(debug_on) masked_blit (dbgbmpt, backbuffer, 0, 0, LMARGIN, 0, SWIDTH-(LMARGIN+RMARGIN), 8);
}

void log_system_info(void)             /* log CPU type, operating system etc */
{
    sprintf(debug_textbuf, "System: cpu %s/%d86.%d/%04x, OS:%08x/%d.%d", 
        cpu_vendor, cpu_family, cpu_model, cpu_capabilities, os_type, os_version, os_revision);
    log_text( 1, debug_textbuf);
}

void check_clwp_errors(void)            /* check for runtime clwp errors */
{   int error = 0;                      /* only in main thread ! */
    
    if(watchdog++ == 200)               /* cleared by keyhandler thread */
    {
        log_text(1, "Watchdog time out");  /* but stuck if no switching... */
        update_help(-1, 9);             /* need to stop... no more IRQ's */
    }
    if (watchdog > 1000) exit(1);       /* hard stop */
//  if (watchdog > 1000) quit = 1;      /* hard stop */
    debug2 = watchdog;                  /* display in debug string */
    
    error = clwpCatch() & 0xFFFF;
    if(error) { log_num(1, "clwpCatch thread: ", clwpCatch() >> 16); }
    if(error) { log_num(1, "clwpCatch err: ", error); }
    clwpCheck();
    error = clwpCatch() & 0xFFFF;
    if(error) { log_num(1, "clwpCatch thread: ", clwpCatch() >> 16); }
    if(error) { log_num(1, "clwpCatch err: ", error); }
    fm_assert(clwpCatch() == 0);
}

/* ---------------------------------------------------------------------
 * Routines for some extras like snow
 * --------------------------------------------------------------------- */


/* check quality of service: if frame rate drops too low, we start
   dropping fancy features, starting with the snow... */
   
int qos[3] = {50, 50, 50};              /* average frame rate for qos */
int qos_sample = 0;

int qos_fish = 1;                       /* quality of service: fish on/off */
int qos_snow = 1;                       /* quality of service: fish on/off */

/* int dummy = 50; code for testing QoS */

void quality_of_service(void)           /* see if we need to reduce features in */
{   int rate;                           /* order to get sufficient frame rates, */
                                        /* e.g. by not displaying snow and/or fish */
    if(++qos_sample & 0x1F) return;     /* don't oversample actual frame rate */
    rate = (qos[0]+qos[1]+qos[2])/3;    /* average frame rate */

/*  if (key[KEY_A]) dummy++;
    if (key[KEY_Z]) dummy--;
    rate = dummy; debug1 = dummy;
    debug2 = rate; 
*/    
    qos[0] = qos[1]; qos[1] = qos[2]; qos[2] = fm_fps;  /* update running average */  
    if((rate < 15) && (qos_snow == 1)) { qos_snow = 0; qos[0]=qos[1]=qos[2]=17; }
    if((rate > 20) && (qos_snow == 0)) { qos_snow = 1; qos[0]=qos[1]=qos[2]=25; }
    if((rate <  7) && (qos_fish == 1) && (qos_snow == 0)) { qos_fish = 0; qos[0]=qos[1]=qos[2]=14; }
    if((rate > 12) && (qos_fish == 0)) { qos_fish = 1; qos[0]=qos[1]=qos[2]=17; }

    debug3 = qos_snow + 2*qos_fish;
}


typedef struct SNOW {                   /* a snow flake object */
    float x, y, vx, vy;                 /* position and speed */
} SNOW;

#define MAX_SNOW 100                    /* number of snow flakes */

SNOW snow[MAX_SNOW];

void init_snow(void)                    /* initialise snow flake data */
{
    int i;
    for(i=0; i<MAX_SNOW; i++)
    {   snow[i].vx = 0;
        snow[i].vy = rand_n(100)/100.0 + 0.1;
        snow[i].x = rand_n(AWIDTH);
        snow[i].y = rand_n(SHEIGHT/2)+TMARGIN;
    }
} 

void let_it_snow(void)                  /* calculate and draw snow flakes */
{   static int cntr = 0;
    int i;                              /* used as timing delay counter */

    if(!qos_snow) { return; }           /* insufficient speed: no snow */
    
    cntr++;
    for(i=0; i<MAX_SNOW; i++)
    {   snow[i].y += snow[i].vy;        /* update snow flake position */
        if((cntr & 0x07) == 0)
        {   snow[i].x += (rand_n(100)-50) / 20;
            if( snow[i].x < 0 ) snow[i].x = AWIDTH-1;
            if( snow[i].x > AWIDTH ) snow[i].x = 0;
        }
        if(snow[i].y >= waterlevel)
        {   snow[i].y = TMARGIN; snow[i].x = rand_n(AWIDTH); }
        putpixel(backbuffer, snow[i].x+LMARGIN, snow[i].y, makecol(255, 255, 255));
        if(i<MAX_SNOW/4)                /* large flake? */
        {
            putpixel(backbuffer, snow[i].x+LMARGIN+1, snow[i].y+1, makecol(255, 255, 255));
            putpixel(backbuffer, snow[i].x+LMARGIN+1, snow[i].y-1, makecol(255, 255, 255));
            putpixel(backbuffer, snow[i].x+LMARGIN-1, snow[i].y+1, makecol(255, 255, 255));
            putpixel(backbuffer, snow[i].x+LMARGIN-1, snow[i].y-1, makecol(255, 255, 255));
        }  
    }
}

void draw_waves(void)                   /* wave effect */
{
    if(qos_sample & 0x20)
    {
        if(qos_sample & 0x40)           /* shift some blocks below surface */
        {                               /* to and fro */
            blit(backbuffer, backbuffer,   0, 300,   1, 300, 200, 60);
            blit(backbuffer, backbuffer,   1, 420,   0, 420, 200, 60);
            blit(backbuffer, backbuffer, 301, 300, 300, 300, 200, 60);
            blit(backbuffer, backbuffer, 300, 420, 301, 420, 200, 60);
        }
        else
        {
            blit(backbuffer, backbuffer,   1, 300,   0, 300, 200, 60);
            blit(backbuffer, backbuffer,   0, 420,   1, 420, 200, 60);
            blit(backbuffer, backbuffer, 300, 300, 301, 300, 200, 60);
            blit(backbuffer, backbuffer, 301, 420, 300, 420, 200, 60);
        }
    }
}

/* ---------------------------------------------------------------------
 * Routines for fish movement and drawing
 * --------------------------------------------------------------------- */

/* copied from other program of mine, so some 'strange' things
    like using fixed point instead of floats */
    
typedef int fix24;                      /* typedef gives better error checking? */

#define I2F(x) ((int)((x)<<8))          /* int to fixed */
#define F2I(x) ((fix24)((x)>>8))        /* fixed to int */

#define X2S(a) (((a)-mapoff.x+20*SWIDTH) % SWIDTH)  /* wrap fish on wide world */
#define Y2S(a) ((a)-mapoff.y)

#define FISHES 20
        
typedef struct FISH {                   /* a fish object */
    fix24 x, y, vx, vy;
    int state;
} FISH;

FISH fishes[FISHES];

/* below structures etc define the potential field, this field is generated
   by objects and might be repellant (positive) or atractive (negative).
   In order to reduce calculation time, the resolution is much lower than
   the actual playing field, by a factor of 2^SCALE in each dimension.
   Only half of array is used, with offset POTSIZE/2; this way I don't
   have to check for boundary conditions due to overlap with the ????_field
   in case e.g. x=0 (then field runs from -2 to +2)     */

#define FSCALE 4                        /* scale factor 16 display -> field */
#define FSIZE  5                        /* size of field definition matrix */
#define POTSIZE ((SWIDTH>>FSCALE) * (SHEIGHT>>FSCALE))
#define XY2POT(x,y) ((y)*(SWIDTH>>FSCALE) + (x) + POTSIZE/2)

int potential[POTSIZE*2];               /* field generated by objects */

int pinguin_field[FSIZE*FSIZE] =        /* pinguins repel fish */
    {   1, 2, 4, 2, 1,                  /* positive = repellant */
        2, 6, 8, 6, 2,                  /* note: scaling in init_potentials() */
        4, 8, 9, 8, 4,
        2, 6, 8, 6, 2,
        1, 2, 4, 2, 1 };

int fish_field[FSIZE*FSIZE] =           /* attract if not too close */
    {   -1, -4, -2, -4, -1,
        -5,  0,  4,  0, -3,
        -3,  4,  8,  4, -5,
        -5,  0,  4,  0, -3,
        -1, -4, -2, -4, -1 };
        

void init_potentials(void)              /* initialise indiv. fields */
{   int k; 
    
    for(k=0; k<FSIZE*FSIZE; k++)
    {
        pinguin_field[k] *= 200;         /* scale up by factor */
        fish_field[k] *= 2;
    }
}

void update_field(fix24 x, fix24 y, int *field)
{   int k, l, ix, iy;

    ix = F2I(x); iy = F2I(y);           /* fixed to int */

    fm_assert(iy > -SHEIGHT/2 + 8);
    fm_assert(iy < SHEIGHT + SHEIGHT/2 - 8);

    for (l=0; l<FSIZE; l++)
        for (k=0; k<FSIZE; k++)
            potential[XY2POT( (ix>>FSCALE)+k-(FSIZE-1)/2, (iy>>FSCALE)+l-(FSIZE-1)/2)]
                += *(field++);
}

void update_potential(void)             /* generate pot. field */
{   int k;
    int tempx, tempy;
    
    for(k=0; k<POTSIZE*2; k++)
        potential[k]=0;                 /* first init potential field at zero */
//    update_field(X2S((int)player.x), Y2S((int)player.y), pinguin_field);

    tempx = ( ((int)player.x) % SWIDTH) + LMARGIN; tempy = player.y + TMARGIN;
    update_field(I2F(tempx), I2F(tempy), pinguin_field);
    
    for(k=0; k<FISHES; k++)             /* update with all fish */
        update_field(fishes[k].x, fishes[k].y, fish_field);
}

int get_gradient(fix24 x, fix24 y, int flag, int *field)
{
    if(!flag)                           /* 0: x gradient */
        return potential[XY2POT( F2I(x)>>FSCALE,   F2I(y)>>FSCALE)]
             - potential[XY2POT( (F2I(x)>>FSCALE)+1, F2I(y)>>FSCALE)]
             - 7;                       /* compensate own field */
    else                                /* 1: y gradient */
        return potential[XY2POT( F2I(x)>>FSCALE, F2I(y)>>FSCALE)]
             - potential[XY2POT( F2I(x)>>FSCALE, (F2I(y)>>FSCALE)+1)]
             - 7;                       /* why not 8 ?????? */
}

void init_fish(void)                    /* initialise start positions etc */
{   int k;
    
    init_potentials();
    for(k=0; k<FISHES; k++)
    {   fishes[k].x = I2F( (rand()%SWIDTH) *7/8 );
        fishes[k].y = I2F( (rand()%SHEIGHT)*3/8 + SHEIGHT/2 );
        fishes[k].vx = rand() % 100 - 50; fishes[k].vy = 0;
        fishes[k].state = FM_FISH;
    }
}

void update_fish(void)                  /* acc. to gradient */
{
    int i;
    
    if(!qos_fish) { return; }           /* insufficient speed: drop fish */
    
    update_potential();                 /* for both player and fish */
    
    for(i=0; i<FISHES; i++)
    {                                   /* calc gradient, update acc. */
        if( (abs(fishes[i].vx) + abs(fishes[i].vy)) > 300)
        {
            fishes[i].vx = (fishes[i].vx * 3 / 4);  /* slow down */
            fishes[i].vy = (fishes[i].vy * 3 / 4);
        }
        fishes[i].vx
            += get_gradient(fishes[i].x, fishes[i].y, 0, fish_field)
            +  (fix24)(rand()%11-5);
            
        fishes[i].vy
            += get_gradient(fishes[i].x, fishes[i].y, 1, fish_field)
            +  (fix24)(rand()%11-5);

        fishes[i].x += fishes[i].vx;    /* update position */
        fishes[i].y += fishes[i].vy;    /* based on speed */
        
        if( F2I(fishes[i].x) < 8)       /* limit to about screen size */
        {   fishes[i].vx += 4;
            if( F2I(fishes[i].x) < -40) fishes[i].x = I2F(-40);
        }
        if( F2I(fishes[i].y) < waterlevel+60)
        {   fishes[i].vy += 8;          /* don't jump above the water level */
            if( F2I(fishes[i].y) < waterlevel+20)
            { fishes[i].y = I2F(waterlevel+20); fishes[i].vy = 1000; }
        }
        if( F2I(fishes[i].x) > SWIDTH-40)
        {   fishes[i].vx -= 4;
            if( F2I(fishes[i].x) > SWIDTH+40) fishes[i].x = I2F(SWIDTH+40);
        }
        if( F2I(fishes[i].y) > SHEIGHT-40)
        {   fishes[i].vy -= 4;
            if( F2I(fishes[i].y) > SHEIGHT+40) fishes[i].y = I2F(SHEIGHT+40);
        }
        
        if( F2I(fishes[i].y) > SHEIGHT+190)
        { allegro_message("Strange value: x=%d, v=%d; y=%d, v=%d for %d",
               F2I(fishes[i].x), F2I(fishes[i].vx),
               F2I(fishes[i].y), F2I(fishes[i].vy), i);
          fm_error("Stopped in updated_fish");
        }
    }
}

void draw_fish(void)
{
    int k;
    
    if(!qos_fish) { return; }           /* insufficient speed: drop fish */
    for(k=0; k<FISHES; k++)
    {   
        if(fishes[k].vx >= 0)           /* check which direction to swim */
            draw_sprite(backbuffer,     /* left to right */
                (BITMAP *)data[FM_FISH+((k<4)?1:0)].dat, // fishes[k].state].dat,
                X2S(F2I(fishes[k].x)), Y2S(F2I(fishes[k].y)));
        else
            draw_sprite_h_flip(backbuffer,  /* right to left: flip */
                (BITMAP *)data[FM_FISH+((k<4)?1:0)].dat, // fishes[k].state].dat,
                X2S( F2I(fishes[k].x)), Y2S(F2I(fishes[k].y)));
    }
}

/* end of file */
